# Copyright 2020 Akamai Technologies, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from typing import NoReturn


class UnavailableModule:
    """
    Fake module that can be imported but raises when used.
    """

    def __init__(self, module: str) -> None:
        self.__module = module

    def __getattr__(self, item: str) -> NoReturn:
        raise RuntimeError(f"{self.__module} cannot be imported.")


try:
    import numpy
except ImportError:
    numpy = UnavailableModule("numpy")

try:
    import pandas
except ImportError:
    pandas = UnavailableModule("pandas")
